from tqdm import tqdm

from corpus_io.morphemes import Morpheme
from corpus_io.nikl.abc import TOKENIZED_SENT_GEN_TYPE

SENT_MORPHEME_SEP = "__SEN|||MOR__"
MORPHEME_SEP = "__ML|||MR__"


def convert_morpheme_stream_to_writable_format(morpheme_stream):
    for sentence, morphemes in morpheme_stream:
        morpheme_str = MORPHEME_SEP.join([morpheme.to_writable_format() for morpheme in morphemes])
        yield f"{sentence}{SENT_MORPHEME_SEP}{morpheme_str}\n"


def convert_line_into_morphemes(line):
    [sent, morphemes_str] = line.split(SENT_MORPHEME_SEP)
    morpheme_strs = morphemes_str.split(MORPHEME_SEP)
    morphemes = [Morpheme.create_from_formatted_str(morpheme_str) for morpheme_str in morpheme_strs]
    return sent, morphemes


class MorphemeIOService(object):
    @staticmethod
    def write_stream(output_file_path, morpheme_stream: TOKENIZED_SENT_GEN_TYPE):
        str_stream = convert_morpheme_stream_to_writable_format(morpheme_stream)
        with open(output_file_path, 'w') as f:
            for line in tqdm(str_stream):
                f.write(line)

    @staticmethod
    def read_as_stream(file_path):
        with open(file_path, 'r') as f:
            for line in f:
                line = line[:-1]
                if len(line) == 0:
                    break
                yield convert_line_into_morphemes(line)



import abc
from inspect import isabstract
from typing import Iterator, Tuple, List

from corpus_io.commons.json_services import JsonLoadService
from corpus_io.morphemes import Morpheme
from corpus_io.nikl.commons.mappers import extract_sentences_as_stream_from, tokenize_to_morphemes

TOKENIZED_SENT_GEN_TYPE = Iterator[Tuple[str, List[Morpheme]]]


class NIKLTokenizeIOService(abc.ABC):
    registered_subclasses = {}

    def __init_subclass__(cls):
        super().__init_subclass__()
        if not isabstract(cls):
            cls.registered_subclasses[cls.code()] = cls

    @classmethod
    def get_tokenized_sentence_generator(cls, dir_path, max_file_num=None, seed=0) -> TOKENIZED_SENT_GEN_TYPE:
        file_stream = JsonLoadService.load_files_as_stream(dir_path, max_file_num, seed)
        sentence_stream = extract_sentences_as_stream_from(file_stream)
        morpheme_stream = tokenize_to_morphemes(sentence_stream)
        return morpheme_stream

    @classmethod
    @abc.abstractmethod
    def code(cls) -> str:
        raise NotImplementedError

    @staticmethod
    def get_io_service_using_code(code):
        return NIKLTokenizeIOService.registered_subclasses[code]



from dataclasses import dataclass

FIELD_SEP = "__FL|||FR__"


@dataclass(frozen=True)
class Morpheme:
    form: str
    tag: str
    start: int
    end: int

    @classmethod
    def create_from_token(cls, token):
        return cls(form=token.form, tag=token.tag, start=token.start, end=token.end)

    @classmethod
    def create_from_formatted_str(cls, formatted_str):
        [form, tag, start, end] = formatted_str.split(FIELD_SEP)
        return cls(form=form, tag=tag, start=int(start), end=int(end))

    def to_writable_format(self):
        return FIELD_SEP.join([self.form, self.tag, str(self.start), str(self.end)])



from typing import Iterable
from kiwipiepy import Kiwi

from corpus_io.morphemes import Morpheme

_kiwi = Kiwi()


def extract_sentence_as_stream(json_data: dict):
    paragraphs = [news["paragraph"] for news in json_data["document"]]
    for paragraph in paragraphs:
        for sentence in paragraph:
            yield sentence["form"]


def extract_sentences_as_stream_from(jsons: Iterable[dict]):
    for json_data in jsons:
        for sentence in extract_sentence_as_stream(json_data):
            yield sentence


def tokenize_to_morphemes(sentences: Iterable[str]):
    for sentence in sentences:
        tokens = _kiwi.tokenize(sentence)
        morphemes = [Morpheme.create_from_token(token) for token in tokens]
        yield sentence, morphemes



from dataclasses import dataclass
from typing import Dict, List, Optional

from corpus_io.file_io.morphemes_io import MorphemeIOService
from corpus_io.nikl.abc import NIKLTokenizeIOService


@dataclass(frozen=True)
class CorpusMergeOption:
    dir_path: str
    max_file_num: Optional[int] = None


def merge_streams(streams):
    for stream in streams:
        for item in stream:
            yield item


class CorpusTokenizeMergeUseCase(object):
    @staticmethod
    def execute(output_file_path, read_options: Dict[str, List[CorpusMergeOption]], seed=0):
        merged_stream = CorpusTokenizeMergeUseCase.create_corpus_read_stream(read_options, seed)
        MorphemeIOService.write_stream(output_file_path, merged_stream)

    @staticmethod
    def create_corpus_read_stream(read_options: Dict[str, List[CorpusMergeOption]], seed=0):
        streams = CorpusTokenizeMergeUseCase._convert_options_into_streams(read_options, seed)
        return merge_streams(streams)

    @staticmethod
    def _convert_options_into_streams(read_options, seed):
        streams = []
        for code, options in read_options.items():
            io_service_cls = NIKLTokenizeIOService.get_io_service_using_code(code)
            for option in options:
                stream = io_service_cls.get_tokenized_sentence_generator(option.dir_path,
                                                                         max_file_num=option.max_file_num, seed=seed)
                streams.append(stream)
        return streams



import os
import json
import random
from typing import List


class JsonLoadService(object):
    @staticmethod
    def load_file(filepath) -> dict:
        with open(filepath) as f:
            loaded_json = json.load(f)
        return loaded_json

    @staticmethod
    def load_files(dir_path, *args, **kwargs) -> List[dict]:
        return [data for data in JsonLoadService.load_files_as_stream(dir_path, *args, **kwargs)]

    @staticmethod
    def load_files_as_stream(dir_path, max_file_num=None, seed=0):
        json_filenames = sorted([filename for filename in os.listdir(dir_path) if filename.endswith('.json')])

        if max_file_num and max_file_num < len(json_filenames):
            random.seed(seed)
            json_filenames = random.sample(json_filenames, max_file_num)

        for json_filename in json_filenames:
            filepath = os.path.join(dir_path, json_filename)
            yield JsonLoadService.load_file(filepath)
